#!/usr/bin/perl

# Search Keyphrases Analyzer
# Copyright (C) 2003 Robert Boeck
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$version = "v0.91";
$program = "Search Keyphrases Analyzer $version";
$copyright = "Copyright (C) 2003 Robert Boeck";

$infile = '';
$sstr = '';
$se = '';
$begindate = '';
$enddate = '';
$unknown = 0;
$undecoded = 0;
$chronological = 0;
$ignoreimg = 0;
$ignorenonhtml = 0;

%hits = ();
%hits1 = ();
%hits2 = ();
%seu = ();
%udrefs = ();

@refs = ();

$os = $^O;      # detect operating system

&print_copyright;

&getparams;

open (IN, $infile) or die "File $infile not found!\n";
while (<IN>) {
  $line = $_;
  $line =~ /^([^\s]+) [^\s]+ [^\s]+ \[([^\s]+) [^\s]+\] \"[^\s]+ ([^\s]+) [^\"]+\" ([\d|-]+) [\d|-]+ \"([^\"]*)\" \"([^\"]*)\".*$/;
  $address = $1;
  $date = $2;
  $begindate = $date unless ($begindate);
  $file = $3;
  $ref = $5;
  $ref =~ s/^http\:\/\///;
  next if (&is_nohtml ($file) && $ignorenonhtml);
  &check_se;
  if ($se) {
    $sestr = "'$sstr' ($se)";
    # $sestr .= " Ref: $ref" if (! $sstr);
    $udrefs{"$se: $ref"} ++ if ((! $sstr) && $undecoded);
    if ($chronological) {
      print &convert_charset_out (sprintf ("%15s $date \"$file\" $sestr\n", $address));
      # print &convert_charset_out ("$address $date \"$file\" $sestr\n");
      # print &convert_charset_out ("$address\t$date\t\"$file\"\t$sestr\n");
    }
    else {
      $hits{$file} ++;
      $hits1{$file}{$sestr} ++;
      $seu{$se} ++;
      $hits2{$se}{$sstr} ++;
    }
  }
  else {
    push (@refs, $ref) if (($ref =~ /\?/) && $unknown);
  }
}
$enddate = $date;
close (IN);

if (! $chronological) {

  print "\n\nsearch phrases and search engine statistics per file\n";
  print "from $begindate to $enddate\n";
  # @fkeys = reverse sort { $hits{$a} <=> $hits{$b} } keys %hits;
  @fkeys = sort { $hits{$b} <=> $hits{$a} || lc($a) cmp lc($b) } keys %hits;
  foreach $fkey (@fkeys) {
    # @skeys = reverse sort { $hits1{$fkey}{$a} <=> $hits1{$fkey}{$b} } keys %{$hits1{$fkey}};
    @skeys = sort { $hits1{$fkey}{$b} <=> $hits1{$fkey}{$a} || $a cmp $b } keys %{$hits1{$fkey}};
    print &convert_charset_out ("\n$fkey: $hits{$fkey} hits:\n");
    $count = 0;
    foreach $skey (@skeys) {
      $count += $hits1{$fkey}{$skey};
    }
    foreach $skey (@skeys) {
      print &convert_charset_out (sprintf ("%6u * (%5.2f%%) %s\n", $hits1{$fkey}{$skey}, $hits1{$fkey}{$skey}*100/$count, $skey));
    }
  }

  $count = 0;
  foreach $fkey (@fkeys) {
    $count += $hits{$fkey};
  }
  print "\n\nhits from search engines: $count total\n\n";
  foreach $fkey (@fkeys) {
    print &convert_charset_out (sprintf ("%6u * (%5.2f%%) %s\n", $hits{$fkey}, $hits{$fkey}*100/$count, $fkey));
  }

  print "\n\nsearch phrases statistics per search engine\n";
  print "from $begindate to $enddate\n";
  # @sekeys = reverse sort { $seu{$a} <=> $seu{$b} } keys %seu;
  @sekeys = sort { $seu{$b} <=> $seu{$a} || lc($a) cmp lc($b) } keys %seu;
  foreach $sekey (@sekeys) {
    # @stkeys = reverse sort { $hits2{$sekey}{$a} <=> $hits2{$sekey}{$b} } keys %{$hits2{$sekey}};
    @stkeys = sort { $hits2{$sekey}{$b} <=> $hits2{$sekey}{$a} || $a cmp $b } keys %{$hits2{$sekey}};
    print &convert_charset_out ("\n$sekey: $seu{$sekey} hits:\n");
    $count = 0;
    foreach $stkey (@stkeys) {
      $count += $hits2{$sekey}{$stkey};
    }
    foreach $stkey (@stkeys) {
      print &convert_charset_out (sprintf ("%6u * (%5.2f%%) %s\n", $hits2{$sekey}{$stkey}, $hits2{$sekey}{$stkey}*100/$count, $stkey));
    }
  }

  $count = 0;
  foreach $sekey (@sekeys) {
    $count += $seu{$sekey};
  }
  print "\n\nsearch engine usage: $count total hits\n\n";
  $indent = 1;
  foreach $sekey (@sekeys) {
    $i = length ($sekey);
    $indent = $i if ($indent < $i);
  }
  $indent++;
  $indent = 20 if ($indent < 20);
  foreach $sekey (@sekeys) {
    print &convert_charset_out (sprintf ("%${indent}s: %6u hits (%5.2f%%)\n", $sekey, $seu{$sekey}, $seu{$sekey}*100/$count));
  }
  
}  

if ($unknown) {
  print "\n\nreferer of potentially unknown search engines:\n\n";
  @refs = sort @refs;
  foreach $ref (@refs) { print &convert_charset_out ("$ref\n"); }
}

if ($undecoded) {
  print "\n\nreferer of unknown search strings from known search engines:\n\n";
  @udrefkeys = sort { $udrefs{$b} <=> $udrefs{$a} || lc($a) cmp lc($b) } keys %udrefs;
  foreach $udref (@udrefkeys) {
    ($se, $ref) = split(/: /, $udref);
    print &convert_charset_out (sprintf ("%6u * %s (%s)\n", $udrefs{$udref}, $ref, $se));
  }
}


sub check_se {
  # if ($ref =~ /images\.google\./) {
  if ($ref =~ /google\.[^\/]+\/(images|imgres)/) {
    $sstr = &getparam ($ref, "prev=");
    $sstr = $ref if (! $sstr);
    $sstr = &getparam ($sstr, "q=");
    $se = "Google Images";
  }
  elsif ($ref =~ /directory\.google\./) {
    $sstr = $ref;
    $sstr =~ s/directory\.google\.[^\/]+//;
    $se = "Google Directory";
  }
  elsif ($ref =~ /dir\.yahoo\./) {
    $sstr = $ref;
    $sstr =~ s/^.*dir\.yahoo\.[^\/]+//;
    $se = "Yahoo Directory";
  }
  elsif ($ref =~ /yahoo\./) {          # must be before google (google.yahoo.com)
    $sstr = &getparam ($ref, "(p|va)=");
    $se = "Yahoo";
  }
  elsif ($ref =~ /google\./) {
    $sstr = &getparam ($ref, "(p|q)=");
    $sstr = &getparam ($ref, "as_q=") if (! $sstr);
    $sstr = &getparam ($ref, "as_epq=") if (! $sstr);
    $se = "Google";
  }
  elsif ($ref =~ /search\.([^\.\s]+\.)?msn\./) {
    $sstr = &getparam ($ref, "q=");
    $sstr = &getparam ($ref, "MT=") if (! $sstr);
    $se = "MSN Search";
  }
  elsif ($ref =~ /brisbane\.t-online\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "T-Online";
  }
  elsif ($ref =~ /hotbot\./) {         # must be before lycos (hotbot.lycos.de)
    $sstr = &getparam ($ref, "query=");
    $se = "Hotbot";
  }
  elsif ($ref =~ /lycos\.[^\/]+\/dir\//) {
    $sstr = $ref;
    $sstr =~ s/^.*lycos\.[^\/]+\/dir//;
    $se = "Lycos Directory";
  }
  elsif ($ref =~ /lycos\./) {
    $sstr = &getparam ($ref, "query=");
    $se = "Lycos";
  }
  elsif ($ref =~ /sucheaol\.aol\.de\/suche\/bilder\//) {
    $sstr = &getparam ($ref, "q=");
    $se = "AOL Search Image";
  }
  elsif ($ref =~ /suche(|aol)\.aol(|svc)\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "AOL Search";
  }
  elsif ($ref =~ /search\.aol\.c/) {
    $sstr = &getparam ($ref, "query=");
    $se = "AOL Search";
  }
  elsif ($ref =~ /altavista\.[^\/]+\/image\//) {
    $sstr = &getparam ($ref, "q=");
    $se = "AltaVista Image";
  }
  elsif ($ref =~ /altavista\.[^\/]+\/audio\//) {
    $sstr = &getparam ($ref, "q=");
    $se = "AltaVista Audio";
  }
  elsif ($ref =~ /altavista\.[^\/]+\/video\//) {
    $sstr = &getparam ($ref, "q=");
    $se = "AltaVista Video";
  }
  elsif ($ref =~ /altavista\.[^\/]+\/dir\//) {
    $sstr = &getparam ($ref, "q=");
    $se = "AltaVista Directory";
  }
  elsif ($ref =~ /altavista\.[^\/]+\/news\//) {
    $sstr = &getparam ($ref, "q=");
    $se = "AltaVista News";
  }
  elsif ($ref =~ /altavista\./) {
    $sstr = &getparam ($ref, "q=");
    $sstr = &getparam ($ref, "aqa=") if (! $sstr);
    $sstr = &getparam ($ref, "aqp=") if (! $sstr);
    $sstr = &getparam ($ref, "aqo=") if (! $sstr);
    $se = "AltaVista";
  }
  elsif ($ref =~ /alltheweb\.com/) {
    $sstr = &getparam ($ref, "q(|uery)=");
    $se = "alltheweb";
  }
  elsif ($ref =~ /search\.bluewin\./) {
    $sstr = &getparam ($ref, "qry=");
    $se = "Bluewin";
  }
  elsif ($ref =~ /suche\d?\.web\.de/) {
    $sstr = &getparam ($ref, "su=");
    $se = "Web DE";
  }
  elsif ($ref =~ /rubriken\.fireball\.de/) {
    $sstr = $ref;
    $sstr =~ s/rubriken\.fireball\.de//;
    $se = "Fireball Rubriken";
  }
  elsif ($ref =~ /suche\.fireball\.de/) {
    $sstr = &getparam ($ref, "q=");
    $se = "Fireball";
  }
  elsif ($ref =~ /netscape\./) {
    $sstr = &getparam ($ref, "q(|uery)=");
    $se = "Netscape";
  }
  elsif ($ref =~ /meta\.ger/) {
    $sstr = "";
    $se = "MetaGer";
  }
  elsif ($ref =~ /evreka\.passagen\.se/) {
    $sstr = &getparam ($ref, "q=");
    $se = "Evreka";
  }
  elsif ($ref =~ /metacrawler\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "Metacrawler";
  }
  elsif ($ref =~ /buscador\.terra\./) {
    $sstr = &getparam ($ref, "query=");
    $se = "Terra";
  }
  elsif ($ref =~ /ask\.com/) {
    $sstr = &getparam ($ref, "ask=");
    $se = "Ask Jeeves";
  }
  elsif ($ref =~ /abacho\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "Abacho";
  }
  elsif ($ref =~ /teoma\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "Teoma";
  }
  elsif ($ref =~ /search\.dogpile\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "Dogpile";
  }
  elsif ($ref =~ /search\.earthlink\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "Earthlink";
  }
  elsif ($ref =~ /suche\.compuserve\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "Compuserve Suche";
  }
  elsif ($ref =~ /suche\.stern\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "Stern Suche";
  }
  elsif ($ref =~ /suchen\.com/) {
    $sstr = &getparam ($ref, "q=");
    $se = "Suchen.com";
  }
  elsif ($ref =~ /austronaut\.at/) {
    $sstr = &getparam ($ref, "suche=");
    $se = "Austronaut";
  }
  elsif ($ref =~ /suchmaschine\.de/) {
    $sstr = &getparam ($ref, "keyword=");
    $se = "suchmaschine.de";
  }
  elsif ($ref =~ /zoek\.vinden\.nl/) {
    $sstr = &getparam ($ref, "query=");
    $se = "vinden.nl";
  }
  elsif ($ref =~ /meinestadt\./) {
    $sstr = &getparam ($ref, "words=");
    $se = "meinestadt.de";
  }
  elsif ($ref =~ /allesklar\./) {
    $sstr = &getparam ($ref, "words=");
    $se = "allesklar.de";
  }
  elsif ($ref =~ /cityreview\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "cityreview.de";
  }
  elsif ($ref =~ /search\-dyn\.excite\./) {
    $sstr = &getparam ($ref, "key=");
    $se = "Excite";
  }
  elsif ($ref =~ /search\-dyn\.tiscali\./) {
    $sstr = &getparam ($ref, "key=");
    $se = "Tiscali";
  }
  elsif ($ref =~ /ixquick\.com/) {
    $sstr = &getparam ($ref, "query=");
    $se = "ixquick";
  }
  elsif ($ref =~ /sapo\.pt/) {
    $sstr = &getparam ($ref, "q=");
    $se = "sapo.pt";
  }
  elsif ($ref =~ /search\.kvasir\.no/) {
    $sstr = &getparam ($ref, "q=");
    $se = "Kvasir";
  }
  elsif ($ref =~ /search\.virgilio\.it/) {
    $sstr = &getparam ($ref, "qs=");
    $se = "Virgilio";
  }
  elsif ($ref =~ /vivisimo\.com/) {
    $sstr = &getparam ($ref, "query=");
    $se = "Vivisimo";
  }
  elsif ($ref =~ /blinde\-kuh\.de/) {
    $sstr = &getparam ($ref, "q=");
    $se = "Blinde Kuh";
  }
  elsif ($ref =~ /caloweb\.de/) {
    $sstr = &getparam ($ref, "q=");
    $se = "caloweb";
  }
  elsif ($ref =~ /qualigo\.de/) {
    $sstr = &getparam ($ref, "q=");
    $se = "QualiGO";
  }
  elsif ($ref =~ /hoogoo\.de/) {
    $sstr = &getparam ($ref, "q=");
    $se = "Hoogoo";
  }
  elsif ($ref =~ /212\.227\.33\.241/) {
    $sstr = &getparam ($ref, "qry=");
    $se = "metaspinner";
  }
  elsif ($ref =~ /^(161\.58\.227\.204|161\.58\.247\.101|212\.40\.165\.90|213\.133\.108\.202|217\.160\.108\.151|217\.160\.111\.99|217\.160\.131\.108|217\.160\.142\.227|217\.160\.176\.42)\/\?/) {
    $sstr = &getparam ($ref, "qry=");
    $se = "metacrawler.de";
  }
  elsif ($ref =~ /metacrawler\.com/) {
    $sstr = &getparam ($ref, "q=");
    $se = "metacrawler.com";
  }
  elsif ($ref =~ /apollo7\.de/) {
    $sstr = &getparam ($ref, "QUERY=");
    $se = "apollo7.de";
  }
  elsif ($ref =~ /flix\.de/) {
    $sstr = &getparam ($ref, "stichwort=");
    $se = "flix.de";
  }
  elsif ($ref =~ /123\.ch/) {
    $sstr = &getparam ($ref, "search=");
    $se = "123.ch";
  }
  elsif ($ref =~ /overture\./) {
    $sstr = &getparam ($ref, "Keywords=");
    $se = "Overture";
  }
  elsif ($ref =~ /gigablast\./) {
    $sstr = &getparam ($ref, "q=");
    $se = "Gibablast";
  }
  elsif ($ref =~ /voila\.com/) {
    $sstr = &getparam ($ref, "kw=");
    $se = "Voila";
  }
  elsif ($ref =~ /suchedeutschland\.com/) {
    $sstr = "";
    $se = "Suche Deutschland";
  }
  elsif ($ref =~ /mirago\.[^\/]+\/scripts\/qhandler/) {
    $sstr = &getparam ($ref, "qry=");
    $se = "mirago";
  }
  elsif ($ref =~ /suche\.freenet\.de/) {
    $sstr = &getparam ($ref, "query=");
    $se = "freenet.de";
  }
  elsif ($ref =~ /scrubtheweb\.com/) {
    $sstr = &getparam ($ref, "(keyword|q)=");
    $se = "Scrub The Web";
  }
  elsif ($ref =~ /(\?|&)(p|q|qr|qt|qry|query|query.?string|qkw|kw|key|keyword|keywords|general|search|string|su|w|word|words|ask|MT|suche)=/) {
    $sstr = &getparam ($ref, "(p|q|qr|qt|qry|query|query.?string|qkw|kw|key|keyword|keywords|general|search|string|su|w|word|words|ask|MT|suche)=");
    $ref =~ /([^\/]*)/;
    $se = "unknown ($1)";
  }
  else {
    $sstr = "";
    $se = "";
  }
  if ($ignoreimg && ($se =~ /Image/)) {
    $se = "";
    $ref = "";
  }
  $sstr = &normalize_searchstr ($sstr) if ($se !~ /(directory|rubriken)/i);
}

sub getparam {
  my $result = shift;
  my $parn = shift;
  $result =~ s/.*\?//;
  my @params = split ("&", $result);
  $result = "";
  my $utf8 = "";
  foreach my $param (@params) {
    if ($param =~ s/^$parn//) {
      $result = $param;
    }
    elsif ($param =~ /^ie=utf-8$/i) {
      $utf8 ++;
    }
  }
  $result = &decodeencodedstring ($result);
  $result = &utf8_to_latin1 ($result); # if ($utf8);
  return $result;
}

sub utf8_to_latin1 {
  my $string = shift;

  my $format = $ENV{"UCFORMAT"}||('%lx');
  use bytes;
  $string =~ s/([\xC0-\xDF])([\x80-\xBF])/sprintf ("%c", hex(sprintf($format,unpack("c",$1)<<6&0x07C0|unpack("c",$2)&0x003F)))/ge;
  $string =~ s/([\xE0-\xEF])([\x80-\xBF])([\x80-\xBF])/sprintf ("%c", hex(sprintf($format,unpack("c",$1)<<12&0xF000|unpack("c",$2)<<6&0x0FC0|unpack("c",$3)&0x003F)))/ge;
  $string =~ s/([\xF0-\xF7])([\x80-\xBF])([\x80-\xBF])([\x80-\xBF])/sprintf ("%c", hex(sprintf($format,unpack("c",$1)<<18&0x1C0000|unpack("c",$2)<<12&0x3F000|unpack("c",$3)<<6&0x0FC0|unpack("c",$4)&0x003F)))/ge;
  no bytes;

  # use utf8;
  # $string =~ tr/\0-\x{FF}//UC;
  # no utf8;

  return $string;
}

sub decodeencodedstring {
  my $stringtodecode = shift;
  $stringtodecode =~ tr/\+/ /s;
  $stringtodecode =~ s/%([A-F0-9][A-F0-9])/pack("C", hex($1))/ieg;
  $stringtodecode =~ s/\\x([A-F0-9][A-F0-9])/pack("C", hex($1))/ieg;
  return $stringtodecode;
}

sub normalize_searchstr {
  my $string = shift;
  $string = lc ($string);        # to lower case
  $string =~ tr///; # special chars to lower case
  $string =~ tr/\+\&\*/ /;       # substitute special characters
  $string =~ s/\s+/ /g;          # shrink whitespace
  ($string) = &trim ($string);   # cut whitespace at beginning and end
  return $string;
}

sub getparams {
  if ($#ARGV < 0) {
    die "usage: ska.pl [-u] [-v] [-i] [-n] <logfile>\n\n"
       ."       -u  show referrers of possibly unknown search engines\n"
       ."       -v  show referrers of unknown search strings from known search engines\n"
       ."       -c  show summary in chronological order\n"
       ."       -i  ignore hits from image searches\n"
       ."       -n  ignore non-html files (.gif, .jpg, .png, .css, .js, .ico)\n";
  }
  foreach my $arg (@ARGV) {
    if ($arg =~ /^\-/) {
      if ($arg =~ /^\-u$/i) { $unknown ++; }
      elsif ($arg =~ /^\-v$/i) { $undecoded ++; }
      elsif ($arg =~ /^\-c$/i) { $chronological ++; }
      elsif ($arg =~ /^\-i$/i) { $ignoreimg ++; }
      elsif ($arg =~ /^\-n$/i) { $ignorenonhtml ++; }
      else { die "unknown parameter: $arg\n"; }
    }
    elsif (! $infile) { $infile = $arg; }
    else { die "unknown parameter: $arg\n"; }
  }
}

sub print_copyright {
  print "\n$program, $copyright\n";
  print "This program is free software and comes with ABSOLUTELY NO WARRANTY;\n";
  print "see the source code for details.\n\n";
}

sub iso_to_ibm850 {
  my $line = shift;

  $line =~ tr//᠅ꤥ净/;

  return $line;
}

sub convert_charset_out {
  my $line = shift;
  if ($os !~ /linux/i) {
    $line = &iso_to_ibm850 ($line);
  }
  return $line;
}

sub trim {
  my ($t, @a);
  foreach $t (@_) {
    $t =~ s/^\s+//;
    $t =~ s/\s+$//;
    push (@a, $t);
  }
  return (@a);
}

sub is_nohtml {
  my $fn = shift;
  my $result = $fn =~ /(\.gif|\.jpg|\.png|\.css|\.js|\.ico)$/i;
  return $result;
}
